import numpy as np
from clloader.datasets import MNIST
from scipy import ndimage


class PermutedMNIST(MNIST):
    """A dataset made of MNIST and pixels permutations.

    The first task is the famous MNIST with 10 classes. Then all following tasks
    are the same MNIST but with pixels permuted in a random way.

    Note that classes are the same, only their representation changes.

    # Reference:
        * Overcoming catastrophic forgetting in neural networks
          Kirkpatrick et al.
          PNAS 2017

    :param nb_permutations: Number of permutations in addition of the original MNIST.
    """

    def __init__(self, *args, nb_permutations=4, **kwargs):
        MNIST.__init__(self, *args, **kwargs)

        self._transformations = list(range(nb_permutations))
        self._mapping = None

    @property
    def need_class_remapping(self) -> bool:
        """Flag for method `class_remapping`."""
        return True

    def class_remapping(self, class_ids: np.ndarray) -> np.ndarray:
        """Optional class remapping.

        Remap class ids so that whatever the permutations, the targets stay
        the same (0-9).
        For example, the second task, with permuted pixels, has targets (10-19)
        in order to mimick a class-incremental training but in reality those
        targets are (0-9).
        The remaping is done so that for the end-user, any tasks of PermutedMNIST
        has targets in the (0-9) range.

        :param class_ids: Original class_ids.
        :return: A remapping of the class ids.
        """
        if self._mapping is None:
            self._mapping = np.concatenate(
                [np.arange(10) for _ in range(len(self._transformations) + 1)]
            )
        return self._mapping[class_ids]

    def init(self):
        base_train, base_test = MNIST.init(self)

        x_train, y_train = [base_train[0]], [base_train[1]]
        x_test, y_test = [base_test[0]], [base_test[1]]

        class_increment = len(np.unique(base_train[1]))
        for i, value in enumerate(self._transformations, start=1):
            trsf_train, trsf_test = self._transform(base_train[0], base_test[0], value)

            x_train.append(trsf_train)
            x_test.append(trsf_test)

            y_train.append(base_train[1] + i * class_increment)
            y_test.append(base_test[1] + i * class_increment)

        x_train = np.concatenate(x_train)
        y_train = np.concatenate(y_train)
        x_test = np.concatenate(x_test)
        y_test = np.concatenate(y_test)

        return (x_train, y_train), (x_test, y_test)

    def _transform(self, x_train, x_test, seed):
        random_state = np.random.RandomState(seed=seed)
        permutations = random_state.permutation(x_train.shape[1] * x_train.shape[2])

        train_shape = x_train.shape
        test_shape = x_test.shape

        x_train = x_train.reshape((train_shape[0], -1))[..., permutations].reshape(train_shape)
        x_test = x_test.reshape((test_shape[0], -1))[..., permutations].reshape(test_shape)

        return x_train, x_test


class RotatedMNIST(PermutedMNIST):
    """A dataset made of MNIST and various rotations.

    The first task is the famous MNIST with 10 classes. Then all following tasks
    are the same MNIST but with a fixed rotations per task.

    Note that classes are the same, only their representation changes.

    # Reference:
        * Gradient Episodic Memory for Continual Learning
          Lopez-Paz and Ranzato
          NeurIPS 2017

    :param angles: A list of angles used in the rotation.
    """

    def __init__(self, *args, angles=[45, 90, 135, 180], **kwargs):
        MNIST.__init__(self, *args, **kwargs)

        self._transformations = angles
        self._mapping = None

    def _transform(self, x_train, x_test, angle):
        x_train = ndimage.rotate(x_train, angle=angle, axes=(2, 1), reshape=False)
        x_test = ndimage.rotate(x_test, angle=angle, axes=(2, 1), reshape=False)

        return x_train, x_test



import os
import tempfile

import numpy as np
import pytest
from PIL import Image
from torch.utils.data import DataLoader

from clloader import CLLoader
from clloader.datasets import PermutedMNIST, RotatedMNIST

# yapf: disable


@pytest.mark.parametrize("nb_permutations,nb_tasks", [
    (0, 1),
    (4, 5)
])
def test_permuted(tmp_path_factory, nb_permutations, nb_tasks):
    dataset = PermutedMNIST(
        data_path=tmp_path_factory.getbasetemp(),
        download=True,
        nb_permutations=nb_permutations
    )

    clloader = CLLoader(dataset, increment=10)

    assert clloader.nb_tasks == nb_tasks
    seen_tasks = 0
    for task_id, (train_dataset, test_dataset) in enumerate(clloader):
        seen_tasks += 1

        for _ in DataLoader(train_dataset):
            pass
        for _ in DataLoader(test_dataset):
            pass

        assert np.max(train_dataset.y) == 9
        assert np.min(train_dataset.y) == 0
        assert np.max(test_dataset.y) == 9
        assert np.min(test_dataset.y) == 0
    assert seen_tasks == nb_tasks


@pytest.mark.parametrize("angles,nb_tasks", [
    ([], 1),
    ([45, 180], 3),
    ([45, 90, 135, 180], 5)
])
def test_rotated(tmp_path_factory, angles, nb_tasks):
    dataset = RotatedMNIST(
        data_path=tmp_path_factory.getbasetemp(),
        download=True,
        angles=angles
    )

    clloader = CLLoader(dataset, increment=10)

    assert clloader.nb_tasks == nb_tasks
    seen_tasks = 0
    for task_id, (train_dataset, test_dataset) in enumerate(clloader):
        seen_tasks += 1

        for _ in DataLoader(train_dataset):
            pass
        for _ in DataLoader(test_dataset):
            pass

        assert np.max(train_dataset.y) == 9
        assert np.min(train_dataset.y) == 0
        assert np.max(test_dataset.y) == 9
        assert np.min(test_dataset.y) == 0
    assert seen_tasks == nb_tasks



from typing import Callable, List, Tuple, Union

import numpy as np
import torch
from clloader.datasets import BaseDataset
from clloader.viz import plot
from PIL import Image
from torchvision import transforms


class Dataset(torch.utils.data.Dataset):
    """A task dataset returned by the CLLoader.

    :param x: The data, either image-arrays or paths to images saved on disk.
    :param y: The targets, not one-hot encoded.
    :param trsf: The transformations to apply on the images.
    :param open_image: Whether to open image from disk, or index in-memory.
    """

    def __init__(
        self, x: np.ndarray, y: np.ndarray, trsf: transforms.Compose, open_image: bool = False
    ):
        self.x, self.y = x, y
        self.trsf = trsf
        self.open_image = open_image

    @property
    def nb_classes(self):
        """The number of classes contained in the current task."""
        return len(np.unique(self.y))

    def add_memory(self, x_memory: np.ndarray, y_memory: np.ndarray):
        """Add memory for rehearsal.

        :param x_memory: Sampled data chosen for rehearsal.
        :param y_memory: The associated targets of `x_memory`.
        """
        self.x = np.concatenate((self.x, x_memory))
        self.y = np.concatenate((self.y, y_memory))

    def plot(self, path=None, figsize=None, nb_per_class=5):
        """Plot samples of the current task, useful to check if everything is ok.

        :param path: If not None, save on disk at this path.
        :param figsize: The size of the figure.
        :param nb_per_class: Amount to sample per class.
        """
        plot(self, figsize=figsize, path=path, nb_per_class=nb_per_class)

    def __len__(self):
        """The amount of images in the current task."""
        return self.x.shape[0]

    def get_image(self, index):
        """Returns a Pillow image corresponding to the given `index`.

        :param index: Index to query the image.
        :return: A Pillow image.
        """
        x = self.x[index]
        if self.open_image:
            img = Image.open(x).convert("RGB")
        else:
            img = Image.fromarray(x.astype("uint8"))
        return img

    def __getitem__(self, index):
        """Method used by PyTorch's DataLoaders to query a sample and its target."""
        img = self.get_image(index)
        y = self.y[index]
        img = self.trsf(img)
        return img, y


class CLLoader:
    """Continual Loader, generating datasets for the consecutive tasks.

    :param cl_dataset: A continual dataset.
    :param increment: Either number of classes per task, or a list specifying for
                      every task the amount of new classes.
    :param initial_increment: A different task size applied only for the first task.
                              Desactivated if `increment` is a list.
    :param train_transformations: A list of data augmentation applied to the train set.
    :param common_transformations: A list of transformations applied to both the
                                   the train set and the test set. i.e. normalization,
                                   resizing, etc.
    :param evaluate_on: How to evaluate on val/test, either on all `seen` classes,
                        on the `current` classes, or on `all` classes.
    """

    def __init__(
        self,
        cl_dataset: BaseDataset,
        increment: Union[List[int], int],
        initial_increment: int = 0,
        train_transformations: List[Callable] = None,
        common_transformations: List[Callable] = None,
        evaluate_on="seen"
    ) -> None:
        self.cl_dataset = cl_dataset

        if train_transformations is None:
            train_transformations = []
        if common_transformations is None:
            common_transformations = self.cl_dataset.transformations
        self.train_trsf = transforms.Compose(train_transformations + common_transformations)
        self.test_trsf = transforms.Compose(common_transformations)

        if evaluate_on not in ("seen", "current", "all"):
            raise NotImplementedError(f"Evaluate mode {evaluate_on} is not supported.")
        self.evaluate_on = evaluate_on

        self._setup()

        self.increments = self._define_increments(increment, initial_increment)

    def _define_increments(self, increment: Union[List[int], int],
                           initial_increment: int) -> List[int]:
        if isinstance(increment, list):
            return increment
        increments = []
        if initial_increment:
            increments.append(initial_increment)

        nb_tasks = (self.nb_classes - initial_increment) / increment
        if not nb_tasks.is_integer():
            raise Exception(
                "The tasks won't have an equal number of classes"
                f" with {len(self.class_order)} and increment {increment}"
            )
        increments.extend([increment for _ in range(int(nb_tasks))])

        return increments

    def _setup(self) -> None:
        self.train_data, self.test_data = self.cl_dataset.init()
        unique_classes = np.unique(self.train_data[1])

        self.class_order = self.cl_dataset.class_order or list(range(len(unique_classes)))

    @property
    def nb_classes(self) -> int:
        """Total number of classes in the whole continual setting."""
        return len(np.unique(self.train_data[1]))

    @property
    def nb_tasks(self) -> int:
        """Number of tasks in the whole continual setting."""
        return len(self)

    def __len__(self) -> int:
        """Returns the number of tasks.

        :return: Number of tasks.
        """
        return len(self.increments)

    def __iter__(self):
        """Used for iterating through all tasks with the CLLoader in a for loop."""
        self._counter = 0
        return self

    def __next__(self) -> Tuple[Dataset, Dataset]:
        """An iteration/task in the for loop."""
        if self._counter >= len(self):
            raise StopIteration
        task = self[self._counter]
        self._counter += 1
        return task

    def __getitem__(self, task_index):
        """Returns a task by its unique index.

        :param task_index: The unique index of a task, between 0 and len(loader) - 1.
        :return: A train and test PyTorch's Datasets.
        """
        max_class = sum(self.increments[:task_index + 1])
        min_class = sum(self.increments[:task_index])  # 0 when task_index == 0.

        train = self._select_data(min_class, max_class)
        train_dataset = Dataset(*train, self.train_trsf, open_image=not self.cl_dataset.in_memory)

        # TODO: validation
        if self.evaluate_on == "seen":
            test = self._select_data(0, max_class, split="test")
        elif self.evaluate_on == "current":
            test = self._select_data(min_class, max_class, split="test")
        else:  # all
            test = self._select_data(0, self.nb_classes, split="test")

        test_dataset = Dataset(*test, self.test_trsf, open_image=not self.cl_dataset.in_memory)

        return train_dataset, test_dataset

    def _select_data(self, min_class, max_class, split="train"):
        """Selects a subset of the whole data for a given task.

        :param min_class: The minimum class id.
        :param max_class: The maximum class id.
        :param split: Either sample from the `train` set, the `val` set, or the
                      `test` set.
        :return: A tuple of numpy array, the first item being the data and the
                 second the associated targets.
        """
        if split == "train":
            x, y = self.train_data
        elif split == "val":
            pass  # TODO
        else:
            x, y = self.test_data

        indexes = np.where(np.logical_and(y >= min_class, y < max_class))[0]
        selected_x = x[indexes]
        selected_y = y[indexes]

        if self.cl_dataset.need_class_remapping:
            # A remapping of the class ids is done to handle some special cases
            # like PermutedMNIST or RotatedMNIST.
            selected_y = self.cl_dataset.class_remapping(selected_y)

        return selected_x, selected_y



import abc
from typing import List, Tuple, Union

import numpy as np
from torchvision import datasets as torchdata
from torchvision import transforms

from clloader.datasets import PyTorchDataset


class CIFAR10(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.cifar.CIFAR10, **kwargs)

    @property
    def transformations(self):
        return [
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
        ]


class CIFAR100(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.cifar.CIFAR100, **kwargs)

    @property
    def transformations(self):
        return [
            transforms.ToTensor(),
            transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
        ]


class MNIST(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.MNIST, **kwargs)


class FashionMNIST(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.FashionMNIST, **kwargs)


class KMNIST(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.KMNIST, **kwargs)


class EMNIST(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.EMNIST, **kwargs)


class QMNIST(PyTorchDataset):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type=torchdata.QMNIST, **kwargs)
